package TNA_SL;

import java.util.Random;

/**
 * The Arc class implements an arc in a directed trust graph. In addition to
 * a source and target, the edge is weighted by a Subjective-Logic opinion.
 */
public class Arc implements Comparable<Arc>{
	
	// ************************** PRIVATE FIELDS *****************************
	
	/**
	 * Unique string identifier associated with this Arc
	 */
	private String id_str;
	
	/**
	 * Integer code representing the source node of this Arc
	 */
	private int source;
	
	/**
	 * Integer code representing the target node of this Arc
	 */
	private int target;
	
	/**
	 * Subjective logic Opinion characterizing this trust relationship
	 */
	private Opinion opinion;
	
	
	// *************************** CONSTRUCTORS ******************************
	
	/**
	 * Construct an Arc by providing source & target nodes (w/random Opinion). 
	 * @param src Integer representation of the source node
	 * @param targ Integer representation of the target node
	 */
	public Arc(int src, int targ){
		this.source = src;
		this.target = targ;
		this.id_str = this.print();
		
		Random rand = new Random();
		double u = rand.nextDouble();
		double b = ((1.0-u)/2.0);
		double d = (1.0 - u - b);
		this.opinion = new Opinion(b, d, u, 0.5);
	}
	
	/**
	 * Construct an Arc by providing all fields.
	 * @param src Integer representation of the source node
	 * @param targ Integer representation of the target node
	 * @param id Unique string identifier associated with this arc
	 * @param op Opinion characterizing this trust relationship
	 */
	public Arc(int src, int targ, String id, Opinion op){
		this.id_str = id;
		this.source = src;
		this.target = targ;
		this.opinion = op;
	}
	
	/**
	 * Construct an Arc w/o an Opinion, and a provided id string.
	 * @param src Integer representation of the source node
	 * @param targ Integer representation of the target node
	 * @param id Unique string identifier associated with this arc
	 */
	public Arc(int src, int targ, String id){
		this.id_str = id;
		this.source = src;
		this.target = targ;
		this.opinion = (Opinion) null;
	}
	
	/**
	 * Construct an Arc by providing source, target, and edge Opinion.
	 * @param src Integer representation of the source node
	 * @param targ Integer representation of the target node
	 * @param op SL-Opinion characterizing the (src->target) relationship
	 */
	public Arc(int src, int targ, Opinion op){
		this.source = src;
		this.target = targ;
		this.id_str = this.print();
		this.opinion = op;
	}
	
	/**
	 * Construct an Arc by providing all data fields.
	 * @param src Integer representation of the source node
	 * @param targ Integer representation of the target node
	 * @param op SL-Opinion characterizing the (src->target) relationship
	 * @param id Unique string identifier associated with this arc
	 */
	public Arc(int src, int targ, Opinion op, String id){
		this.source = src;
		this.target = targ;
		this.id_str = id;
		this.opinion = op;
	}
	
	// ************************** PUBLIC METHODS *****************************

	/**
	 * Compare this Arc object to another Arc object. Arcs are compared 
	 * numerically first by the 'source' field, then the 'target' one.
	 * @param other The other Arc to be compared
	 * @return -1 if (this < other), 0 if this equals other; 1 otherwise.
	 */
	public int compareTo(Arc other){		
		int source_compare = this.source - other.source;
		if(source_compare < 0)
			return -1;
		else if(source_compare > 0)
			return 1;
		else{ // Primary comparison is on the 'source' field
			int target_compare = this.target - other.target;
			if(target_compare < 0)
				return -1;
			else if(target_compare > 0)
				return 1;
			else
				return 0;		
		} // If 'source' are equal, compare the 'target' field
	}
	
	/**
	 * Output a string representation of an Arc.
	 * @return A string of the form "[source->target]".
	 */
	public String print(){
		return("["+ this.source + "->" + this.target + "]");
	}
	
	/**
	 * Output a string representation of an Arc, including the edge Opinion
	 * @return A string of the form "[source->target] (b, d, u, a)".
	 */
	public String printWithOp(){
		return("["+ this.source + "->" + this.target + "] " + getOp().print());
	}
	
	/**
	 * Create a duplicate of this Arc object.
	 * @return a duplicate of this Arc object
	 */
	public Arc clone(){
		return (new Arc(this.source, this.target, this.opinion, this.id_str));
	}
	
	// ********************** PUBLIC ACCESS METHODS **************************
	
	/**
	 * Return the 'id_str' field of this Arc.
	 * @return the 'id_str' field of this Arc
	 */
	public String getID(){
		return this.id_str;
	}
	
	/**
	 * Return the 'source' field of this Arc.
	 * @return the 'source' field of this Arc
	 */
	public int getSrc(){
		return this.source;
	}
	
	/**
	 * Return the 'target' field of this Arc.
	 * @return the 'target' field of this Arc
	 */
	public int getTarg(){
		return this.target;
	}
	
	/**
	 * Return the 'opinion' field of this Arc.
	 * @return the 'opinion' field of this Arc
	 */
	public Opinion getOp(){
		return this.opinion;
	}
	
}
