package TNA_SL;

import java.util.*;

/**
 * The Digraph class implements a conventional directed graph of the trust
 * management style, maintaining a collection of Nodes and Arcs.
 */
public class Digraph implements Comparable<Digraph>{
	
	// ************************** PRIVATE FIELDS *****************************
	
	/**
	 * Sorted list containing the Arcs of this Digraph.
	 */
	private List<Arc> arcList;
	 
	/**
	 * Sorted list containing the Nodes of this Digraph.
	 */
	private List<Node> nodeList;
	
	/**
	 * Assuming this Digraph represents a single path, this field describes
	 * the trust confidence of that path (per Subjective-Logic computation).
	 */
	private double confidence;
	
	// *************************** CONSTRUCTORS ******************************
	
	/**
	 * Construct an empty Digraph object.
	 */
	public Digraph(){
		this.confidence = 0.0;
		this.arcList = new ArrayList<Arc>();
		this.nodeList = new ArrayList<Node>();
	}
	
	// ************************** PUBLIC METHODS *****************************

	/**
	 * Given the fact that this Digraph is a Direct-Series-Parallel-Graph,
	 * will the addition of the Arcs in 'branch' keep it a DSPG?
	 * @param src Integer identifying source terminal of this DSPG
	 * @param targ Integer identifying sink terminal of this DSPG
	 * @param branch Digraph of the Arcs to be added
	 * @return True if (this+branch) is a DSPG, false otherwise
	 */
	public boolean sep_subgraph(int src, int targ, Digraph branch){

		  // Make a graph copy and add branch nodes to it
		Digraph copy = this.clone();
		for(int i=0; i < branch.arcSize(); i++)
			copy.addArc(branch.arcGet(i));
			
		List<Node> unsatisfied = new ArrayList<Node>();
		for(int i=0; i < copy.nodeSize(); i++){
			unsatisfied.add(copy.nodeGet(i));
		} // Make list of unsatisfied nodes, add all nodes to it
		
		boolean changeMade = true;	// Indicates if reduction is progressing
		int new_src, new_targ;		// Source and target of replacement Arc
		Node cur_node;				// Pointer to Node being worked on
		Arc new_arc;				// Replacement arc as result of reduction
		
		while(changeMade){
			changeMade = false;
			for(int i=0; i < unsatisfied.size(); i++){
				cur_node = unsatisfied.get(i);
				if(cur_node.inSize() == 1 && cur_node.outSize() == 1){
					
						// Build up the replacement Arc
					new_src = cur_node.inGetArc(0).getSrc();
					new_targ = cur_node.outGetArc(0).getTarg();
					new_arc = new Arc(new_src, new_targ, "");
					
						// Remove all instances of previous, add new Arc
					copy.removeArc(cur_node.inGetArc(0));
					copy.removeArc(cur_node.outGetArc(0));
					copy.addArc(new_arc);
					unsatisfied.remove(i);
					changeMade = true;
				} // If only one arc in and out; do series reduction
			
				if(changeMade)
					break; // Due to possible index shift above; restart
				
				if(cur_node.outSize() >= 2 && cur_node.outGetArc(0).getTarg() 
						== cur_node.outGetArc(1).getTarg()){

						// Build up the replacement Arc
					new_src = cur_node.outGetArc(0).getSrc();
					new_targ = cur_node.outGetArc(0).getTarg();
					new_arc = new Arc(new_src, new_targ, "");
					
						// Remove all instances of previous, add new Arc
					copy.removeArc(cur_node.outGetArc(0)); // Shifts index
					copy.removeArc(cur_node.outGetArc(0));
					copy.addArc(new_arc);
					changeMade = true;
				} // If two outgoing Arcs have same target; parallel reduction
				
			} // Each iteration check all unsatisfied nodes for reduction
		} // Keep reducing so long as each pass makes a single update
		
		boolean isDSPG = true;
		for(int i=0; i < copy.arcSize(); i++){
			if(copy.arcGet(i).getSrc()!=src || copy.arcGet(i).getTarg()!=targ)
				isDSPG = false;
		} // If any arcs remain not between source and target, no DSPG
		return isDSPG;
	}
	
	/**
	 * Given the fact that this Digraph is a Direct-Series-Parallel-Graph,
	 * perform the reduction and compute the single Opinion that results.
	 * @param src Integer identifying source terminal of this DSPG
	 * @param targ Integer identifying target terminal of this DSPG
	 * @return Opinion characterizing the reduced DSPG
	 */
	public Opinion dspgCompute(int src, int targ){
		
		Digraph copy = this.clone();			
		List<Node> unsatisfied = new ArrayList<Node>();
		for(int i=0; i < copy.nodeSize(); i++){
			unsatisfied.add(copy.nodeGet(i));
		} // Make list of unsatisfied nodes, add all nodes to it
		
		boolean changeMade = true;	// Indicates if reduction is progressing
		Opinion lhs, rhs, new_op;	// For building up the parse tree
		int new_src, new_targ;		// Source and target of replacement Arc
		Node cur_node;				// Pointer to Node being worked on
		Arc new_arc;				// Replacement arc as result of reduction
		new_op = null;
		
		while(changeMade){
			changeMade = false;
			for(int i=0; i < unsatisfied.size(); i++){
				cur_node = unsatisfied.get(i);
				if(cur_node.inSize() == 1 && cur_node.outSize() == 1){
					
						// Build up the replacement Arc
					new_src = cur_node.inGetArc(0).getSrc();
					new_targ = cur_node.outGetArc(0).getTarg();
					lhs = cur_node.inGetArc(0).getOp();
					rhs = cur_node.outGetArc(0).getOp();
					new_op = lhs.discount(rhs);
					new_arc = new Arc(new_src, new_targ, "", new_op.clone());
					
						// Remove all instances of previous, add new Arcs
					copy.removeArc(cur_node.inGetArc(0));
					copy.removeArc(cur_node.outGetArc(0));
					copy.addArc(new_arc);
					unsatisfied.remove(i);
					changeMade = true;
				} // If only one arc in and out; do series reduction
			
				if(changeMade)
					break; // Due to possible index shift above; restart
				
				if(cur_node.outSize() >= 2 && cur_node.outGetArc(0).getTarg() 
						== cur_node.outGetArc(1).getTarg()){

						// Build up the replacement Arc
					new_src = cur_node.outGetArc(0).getSrc();
					new_targ = cur_node.outGetArc(0).getTarg();
					lhs = cur_node.outGetArc(0).getOp();
					rhs = cur_node.outGetArc(1).getOp();
					new_op = lhs.consensus(rhs);
					new_arc = new Arc(new_src, new_targ, "", new_op.clone());
					
						// Remove all instances of previous, add new Arc
					copy.removeArc(cur_node.outGetArc(0)); // Shifts index
					copy.removeArc(cur_node.outGetArc(0));
					copy.addArc(new_arc);
					changeMade = true;
				} // If two outgoing Arcs have same target; parallel reduction
				
			} // Each iteration check all unsatisfied nodes for reduction
		} // Keep reducing so long as each pass makes a single update
		
			// Given this is a DSPG, there will only be one Arc remaining
			// after reduction and we can simply return its Opinion
		return copy.arcGet(0).getOp(); 
	}
	
	/**
	 * Given that this Digraph represents a path, use the SL-discount operator
	 * to set the 'confidence' field as the confidence we have in this path. If
	 * there are no or multiple paths, behavior is undefined.
	 * @param src Integer code identifying start node of path
	 * @param targ Integer code identifying end node of path
	 */
	public void setConfidence(int src, int targ){
		if(this.arcSize() == 0)
			this.confidence = 0.0;
		else if(arcSize() == 1)
			this.confidence = this.arcGet(0).getOp().getConfidence();
		else{ // confidence calculation is trivial for small paths
			int index = Collections.binarySearch(this.arcList, new Arc(src,-1));
			Opinion op = this.arcGet((index+1)*-1).getOp();
			while(true){
				index = Collections.binarySearch(this.arcList, new Arc(src,-1));
				op = op.discount(this.arcGet((index+1)*-1).getOp());
				src = this.arcGet((index+1)*-1).getTarg();
				if(src == targ)
					break;
			} // Continue following path until target reached
			this.confidence = op.getConfidence();
		} // else if arcList.size() > 1
	}
	
	/**
	 * Compare two Digraphs. Comparison is done on the 'confidence' field.
	 * @param other The other Digraph object being compared to 'this' one
	 * @return -1 if (this < other), 0 if this equals other; 1 otherwise
	 */
	public int compareTo(Digraph other){
		double confidence = this.confidence - other.confidence;
		if(confidence < 0)
			return 1;
		else if(confidence == 0)
			return 0;
		else
			return -1;
	}
	
	/**
	 * Create a duplicate Digraph object.
	 * @return a duplicate Digraph object
	 */
	public Digraph clone(){
		Digraph clone = new Digraph();
		for(int i=0; i < arcSize(); i++)
			clone.addArc(arcGet(i).clone());
		return clone;
	}
	
	/**
	 * Output a string representation of this Digraph; printing all Arcs.
	 * @return a string representation of this Digraph
	 */
	public String print(){
		String retString = "";
		for(int i=0; i < arcSize(); i++)
			retString += i + ": " + arcGet(i).print() + "\n";
		return retString;
	}
	
	/**
	 * Output String describing this Digraph; including Arc/Opinion data
	 * @return a verbose string representation of this Digraph
	 */
	public String printWithOp(){
		String retString = "";
		for(int i=0; i < arcSize(); i++)
			retString += i + ": " + arcGet(i).printWithOp() + "\n";
		return retString;
	}
	
	// ********************* PUBLIC LIST MAINTENANCE *************************

	/**
	 * Check if this Digraph contains a Node with a given label.
	 * @param label Identity of the Node whose existence is being determined
	 * @return True if this Digraph contains node 'label'; false, otherwise
	 */
	public boolean hasNode(int label){
		if(Collections.binarySearch(this.nodeList, new Node(label)) < 0)
			return false;
		else 
			return true;
	}
	
	/**
	 * Add an Arc to this Digraph (done in a sorted fashion).
	 * @param a Arc to be added
	 */
	public void addArc(Arc a){
		
			// Make sure both source and target Node already exist; not a cycle
		if(a.getSrc() == a.getTarg())
			return;
		if(!this.hasNode(a.getSrc()))
			addNode(a.getSrc());
		if(!this.hasNode(a.getTarg()))
			addNode(a.getTarg());
		
			// Insert Arc into arcList in a sorted fashion
		int index = Collections.binarySearch(this.arcList, a);
		if(index >= 0)
			this.arcList.add(index, a);
		else
			this.arcList.add((index+1)*-1, a);
		
			// Add Arc to appropriate Node lists in a sorted fashion
		index = Collections.binarySearch(this.nodeList, new Node(a.getSrc()));
		this.nodeGet(index).outAddArc(a);
		index = Collections.binarySearch(this.nodeList, new Node(a.getTarg()));
		this.nodeGet(index).inAddArc(a);
	}
	
	/**
	 * Remove this Arc from this Digraph.
	 * @param a Arc to be removed
	 */
	public void removeArc(Arc a){	
			// Take this Arc out of the ArcList
		int index = Collections.binarySearch(this.arcList, a);
		this.arcList.remove(index);
		
			// Handle the outgoing list removal
		index = Collections.binarySearch(this.nodeList, new Node(a.getSrc()));
		this.nodeGet(index).outRemArc(a);
		if(nodeIsEmpty(index))
			this.nodeList.remove(index);
		
			// Then do the incoming list removal
		index = Collections.binarySearch(this.nodeList, new Node(a.getTarg()));
		this.nodeGet(index).inRemArc(a);
		if(nodeIsEmpty(index))
			this.nodeList.remove(index);
	}
	
	/**
	 * Determine if this graph contains any Arcs or Nodes.
	 * @return True if this graph contains any data; false, otherwise
	 */
	public boolean isEmpty(){
		if(this.arcList.size() > 0)
			return false;
		else
			return true;
	}
	
	// ******************* TRIVIAL PUBLIC ACCESSOR METHODS *******************

	/**
	 * Retrieve an Arc from the arcList at a specified index.
	 * @param index Index of the Arc to be retrieved
	 * @return Arc at position 'index' in the arcList of this Digraph
	 */
	public Arc arcGet(int index){
		return this.arcList.get(index);
	}
	
	/**
	 * Return the number of Arcs this Digraph contains.
	 * @return the number of Arcs this Digraph contains
	 */
	public int arcSize(){
		return this.arcList.size();
	}
	
	/**
	 * Retrieve the 'confidence' field of this Digraph.
	 * @return the 'confidence' field of this Digraph
	 */
	public double getConfidence(){
		return this.confidence;
	}
	
	/**
	 * Return the number of Nodes this Digraph contains.
	 * @return the number of Nodes this Digraph contains
	 */
	public int nodeSize(){
		return this.nodeList.size();
	}
	
	/**
	 * Retrieve a Node from the nodeList at a specified index.
	 * @param index Index of the Node to be retrieved
	 * @return Node at position 'index' in the nodeList of this Digraph
	 */
	public Node nodeGet(int index){
		return this.nodeList.get(index);
	}
	
	// ************************* PRIVATE METHODS *****************************
	
	/**
	 * Check to see if a Node at a specified index has any connected Arcs.
	 * @param index Index of the Node being examined
	 * @return True if the Node at 'index' has connected Arcs; false, otherwise
	 */
	private boolean nodeIsEmpty(int index){
		boolean isEmpty = (this.nodeGet(index).inSize() == 0);
		isEmpty = (isEmpty && this.nodeGet(index).outSize() == 0);
		return isEmpty;
	}
	
	/**
	 * Add a Node to the nodeList.
	 * @param label Integer identifier of the Node to be added
	 */
	private void addNode(int label){
		Node newNode = new Node(label);
		int index = (((Collections.binarySearch(this.nodeList, newNode))+1)*-1);
		this.nodeList.add(index, newNode);
	}

}
