package TNA_SL;

import java.util.*;

/**
 * The Node class implements a node/vertex in a directed graph.
 */
public class Node implements Comparable<Node>{

	// ************************* PRIVATE FIELDS *****************************
	
	/**
	 * Integer representing the label of this Node.
	 */
	private int id;
	
	/**
	 * List containing all Arcs that are incoming to this Node.
	 */
	private List<Arc> incoming;
	
	/**
	 * List containing all Arcs that are outgoing from this Node.
	 */
	private List<Arc> outgoing;
	
	// *************************** CONSTRUCTORS ******************************
	
	/**
	 * Construct a Node by providing an integer label.
	 * @param label Integer label of the Node to be constructed
	 */
	public Node(int label){
		this.id = label;
		this.incoming = new ArrayList<Arc>();
		this.outgoing = new ArrayList<Arc>();	
	}
	
	// ************************** PUBLIC METHODS *****************************
	
	/**
	 * Add an arc which is incoming to this Node (i.e. 'this' is the target).
	 * @param a The Arc to be added to the incoming list
	 */
	public void inAddArc(Arc a){
		this.incoming.add(a);
	}
	
	/**
	 * Add an arc which is outgoing from this Node (i.e. 'this' is the source).
	 * @param a The Arc to be added to the outgoing list
	 */
	public void outAddArc(Arc a){
		this.outgoing.add(a);
	}
	
	/**
	 * Remove an Arc from this Node's 'incoming' list.
	 * @param a The Arc to be removed from the incoming list
	 * @return True if 'a' was found and removed; false, otherwise
	 */
	public boolean inRemArc(Arc a){
		return this.incoming.remove(a);
	}
	
	/**
	 * Remove an Arc from this Node's 'outgoing' list.
	 * @param a The Arc to be removed from the outgoing list
	 * @return True if 'a' was found and removed; false, otherwise
	 */
	public boolean outRemArc(Arc a){
		return this.outgoing.remove(a);
	}
	
	/**
	 * Determine the number of Arcs in this Node's 'incoming' list.
	 * @return The number of Arcs in the incoming list
	 */
	public int inSize(){
		return this.incoming.size();
	}
	
	/**
	 * Determine the number of Arcs in this Node's 'outgoing' list.
	 * @return The number of Arcs in the outgoing list
	 */
	public int outSize(){
		return this.outgoing.size();
	}
	
	/**
	 * Retrieve an Arc from this Node's 'incoming' list.
	 * @param index Index of the desired Arc object
	 * @return Arc object located at position 'index' in the incoming list
	 */
	public Arc inGetArc(int index){
		return this.incoming.get(index);
	}
	
	/**
	 * Retrieve an Arc from this Node's 'outgoing' list.
	 * @param index Index of the desired Arc object
	 * @return Arc object located at position 'index' in the outgoing list
	 */
	public Arc outGetArc(int index){
		return this.outgoing.get(index);
	}

	/**
	 * Compare this Node object to another Node object.
	 * @param other The other Node to be compared
	 * @return -1 if (this < other), 0 if this equals other; 1 otherwise.
	 */
	public int compareTo(Node other){
		int id_compare = this.id - other.id;
		if(id_compare < 0)
			return -1;
		else if(id_compare > 0)
			return 1;
		else
			return 0;
	}

	/**
	 * Create a duplicate of this Node object.
	 * @return a duplicate of this Node object.
	 */
	public Node clone(){
		Node clone = new Node(this.id);
		for(int i=0; i < this.incoming.size(); i++)
			clone.inAddArc(this.incoming.get(i).clone());
		for(int i=0; i < this.outgoing.size(); i++)
			clone.inAddArc(this.outgoing.get(i).clone());
		return clone;		
	}
	
}
