
package TNA_SL;

import java.util.*;
import interfaces.*;

/**
 * The TNA_Frontend class implements the TMInterface interface, and makes the
 * use of a TNA_SL object (which can stand-alone) more user friendly. Chiefly,
 * it uses strings for user ID's, and permits construction with primitive types.
 */
public class TNA_Frontend implements TMInterface{
	
	// ************************** PRIVATE FIELDS *****************************
	
	/**
	 * Graph to be built; then analyzed by TNA-SL methodology.
	 */
	private Digraph graph;
	
	/**
	 * Map user ID strings (keys), to integers for use by TNA_SL.
	 */
	private Map<String, Integer> userLookup;
	
	// *************************** CONSTRUCTORS ******************************
	
	/**
	 * Construct an empty TNA_Frontend object.
	 */
	public TNA_Frontend(){
		this.graph = new Digraph();
		this.userLookup = new HashMap<String, Integer>();
	}
	
	// ************************ INTERFACED METHODS ***************************
	
	/**
	 * Given a reputation store conforming to RMDBInterface, extract the data
	 * therein, so it may be converted to a form use-able by 'this' object.
	 * Note that this operation clears any previously added data.
	 * @param db Reputation store conforming to RMDBInterface interface
	 */
	public void importRMData(RMDBInterface db){
		int pos, neg;
		this.graph = new Digraph();
		String[] ul = db.getUserList();
		for(int i=0; i < ul.length; i++){
			for(int j=0; j < ul.length; j++){
				pos = db.getPos(ul[i], ul[j]);
				neg = db.getNeg(ul[i], ul[j]);
				this.addEdge(ul[i], ul[j], pos, neg);
			} // Convert reputation database info to TNA-SL format
		} // Use user list to determine all possible relationships
	}
	
	/**
	 * Add a user->user relationship to this TNA_Frontend object.
	 * @param source Unique string ID of the source user in the relationship
	 * @param target Unique string ID of the target user in the relationship
	 * @param pos Number of [source->target] positive interactions
	 * @param neg Number of [source->target] negative interactions
	 */
	public void addEdge(String source, String target, int pos, int neg){
		int sourceInt = getUserIndex(source);
		int targetInt = getUserIndex(target);
		Opinion op = new Opinion(pos, neg);
		graph.addArc(new Arc(sourceInt, targetInt, op));
	}
	
	/**
	 * Compute trust between two parties via TNA_SL methodology
	 * @param source Unique string ID of the source user
	 * @param target Unique string ID of the target user
	 * @return TNA-SL trust of [source->target], or NaN if user(s) are unknown
	 */
	public double computeTrust(String source, String target){
		if(!userExists(source) || !userExists(target))
			return Double.NaN;
		else if(source.equals(target))
			return Double.NaN;
		int sourceInt = getUserIndex(source);
		int targetInt = getUserIndex(target);
		TNA_SL tna = new TNA_SL(this.graph, 0.01, 2); // Suggested parameters
		return (tna.computeTrust(sourceInt, targetInt));
	}
	
	// ************************* PRIVATE METHODS *****************************
	
	/**
	 * Check to see if a user ID is known by this TNA_Frontend
	 * @param userID Unique string ID of some user
	 * @return True if 'userID' is recognized; otherwise, false
	 */
	private boolean userExists(String userID){
		return (this.userLookup.containsKey(userID));
	}
	
	/**
	 * Retrieve the integer index of a String ID; create one if it DNE.
	 * @param userID Unique string ID of some user
	 * @return Integer index associated w/'userID', created or otherwise
	 */
	private int getUserIndex(String userID){
		Integer index = this.userLookup.get(userID);
		if(index == null){
			index = this.userLookup.size();
			this.userLookup.put(userID, index);
		} // If mapping doesn't already exist, create it
		return index;
	}

}
