package TNA_SL;

import java.util.*;
import java.util.regex.Pattern;

/**
 * The TNA_SL class implements trust graph analysis as described by Josang.
 * Given a graph with edges weighted by Opinions, we construct the DSPG that
 * maximizes confidence, and use the SL operators to calculate trust.
 */
public class TNA_SL{
	
	// ************************** PRIVATE FIELDS *****************************
	
	/**
	 * Original trust graph; that passed in by the user
	 */
	private Digraph graph;
	
	/**
	 * The confidence maximized direct-series parallel graph from 'graph'
	 */
	private Digraph dspg;
	
	/**
	 * An enumeration of all non-cyclic paths between two nodes in 'graph'
	 */
	private List<Digraph> paths;
	
	/**
	 * Temporary Graph used in the construction of 'paths'
	 */
	private Digraph path;
	
	/**
	 * A path is no longer considered useful if path.confidence < THRESHOLD
	 */
	private final double THRESHOLD;
	
	/**
	 * Only rejected branches of (magnitude <= MAX_CACHE) are cached
	 */
	private final int MAX_CACHE;
	
	// *************************** CONSTRUCTORS ******************************
	
	/**
	 * Construct a TNA_SL object given a Digraph object.
	 * @param argGraph Graph whose trust will be analyzed
	 * @param threshold Sets 'THRESHOLD' field of this class (0.01 suggested)
	 * @param cacheDepth Sets 'MAX_CACHE' field of this class (2 suggested)
	 */
	public TNA_SL(Digraph argGraph, double threshold, int cacheDepth){
		this.graph = argGraph;
		this.THRESHOLD = threshold;
		this.MAX_CACHE = cacheDepth;
	}
	
	/**
	 * Construct a TNA_SL object given a String graph description.
	 * @param desc A properly formatted trust-graph string. The graph; 
	 * "[1->2,(0.33,0.25,0.42,0.5)]+[2->3,(0.5,0.5,1.0,0.5)]" is an
	 * example of a three-node, two-arc graph. The 4-tuple is the SL-Opinion
	 * characterizing the relationship between two nodes. Nodes must be 
	 * integers. Opinion fields are floating point. The second argument is the
	 * source node (integer) and the third argument the target node (integer).
	 * @param threshold Sets'THRESHOLD' field of this class (0.01 suggested)
	 * @param cacheDepth Sets 'MAX_CACHE' field of this class (2 suggested)
	 */
	public TNA_SL(String desc, double threshold, int cacheDepth){
		this.THRESHOLD = threshold;
		this.MAX_CACHE = cacheDepth;

		String pat = "(\\[\\d+->\\d+,\\s*\\(((1\\.0*)|(0*\\.\\d+)),\\s*" +
				"((1\\.0*)|(0*\\.\\d+)),\\s*((1\\.0*)|(0*\\.\\d+)),\\s*" +
				"((1\\.0*)|(0*\\.\\d+))\\s*\\)\\s*\\])(\\s*\\+\\s*\\[\\d+->" +
				"\\d+,\\s*\\(\\s*((1\\.0*)|(0*\\.\\d+)),\\s*((1\\.0*)|" +
				"(0*\\.\\d+)),\\s*((1\\.0*)|(0*\\.\\d+)),\\s*((1\\.0*)|" +
				"(0*\\.\\d+))\\s*\\)\\s*\\])*";
		
		if(!Pattern.matches(pat, desc)){
			System.err.println("Input graph poorly formatted");
			System.exit(1);
		} else{ // If pattern check fails, exit the program
			pat = "(\\[)|(->)|(,)|(\\+)|(\\()|(\\))|(\\])";
			Pattern p = Pattern.compile(pat);
			String[] raw_data = p.split(desc);
			List<String> data = new ArrayList<String>();
			
			for(int i=0; i < raw_data.length; i++){	
				if(!raw_data[i].isEmpty())
					data.add(raw_data[i]);
			} // Make sure no spaces are left (this has been problematic)
			
			int source, target;
			double b, d, u, a;
			this.graph = new Digraph();
			for(int i=0; i < data.size(); i+=6){
				source = Integer.parseInt(data.get(i));
				target = Integer.parseInt(data.get(i+1));
				b = Double.parseDouble(data.get(i+2));
				d = Double.parseDouble(data.get(i+3));
				u = Double.parseDouble(data.get(i+4));
				a = Double.parseDouble(data.get(i+5));
				graph.addArc(new Arc(source, target, new Opinion(b, d, u, a)));
			} // Then take the Nodes and Opinion data, add to Digraph
		} // Else, we assume the graph is well-formatted and proceed 
		return;
	}
	
	// ************************** PUBLIC METHODS *****************************

	/**
	 * Use TNA-SL methodology to compute asymmetric trust between two parties.
	 * @param source User who is making the trust request
	 * @param target User who 'source' wants to evaluate
	 * @return Expected value of the TNA_SL calculated relationship Opinion
	 */
	public double computeTrust(int source, int target){
		
			// Algorithm timing methods (if desired)
		//long startTime = System.currentTimeMillis();
		//long stopTime = System.currentTimeMillis();
		//double runTime = ((stopTime - startTime)/1000.0);

		this.dspg = new Digraph();
		this.paths = new ArrayList<Digraph>();
		this.path = new Digraph();
		
		findPaths(source, target);
		sortPathsByConfidence(source, target);

		findNearOptimalDSPG(source, target);
		Opinion op = dspg.dspgCompute(source, target);
		return (op.expectedValue());
	}
	
	// ************************* PRIVATE METHODS *****************************
	
	/**
	 * Enumerate all possible paths between a two nodes in 'graph'
	 * @param src Integer label of the source Node
	 * @param target Integer label of the target Node
	 */
	private void findPaths(int src, int target){
		boolean addArc;
		Digraph arcs = new Digraph();
		for(int i=0; i < graph.arcSize(); i++){
			addArc = (graph.arcGet(i).getSrc() == src);
			addArc = (addArc && !path.hasNode(graph.arcGet(i).getTarg()));			
			if(addArc)
				arcs.addArc(graph.arcGet(i));	
		} // Arcs are added if they connect to source; aren't already included
		
		for(int i=0; i < arcs.arcSize(); i++){
			if(arcs.arcGet(i).getTarg() == target){
				path.addArc(arcs.arcGet(i));
				path.setConfidence(src, arcs.arcGet(i).getTarg());
				if(path.getConfidence() > THRESHOLD)
					paths.add(path.clone());
				path.removeArc(arcs.arcGet(i));
			} else if(arcs.arcGet(i).getTarg() != target){
				path.addArc(arcs.arcGet(i));
				path.setConfidence(src, arcs.arcGet(i).getTarg());
				if(path.getConfidence() > THRESHOLD)
					findPaths(arcs.arcGet(i).getTarg(), target);
				path.removeArc(arcs.arcGet(i));
			} // If target reached, conclude this path; else recurse
		} // Exhaustively enumerate all non-cyclic paths
	}
	
	/**
	 * Sort the 'paths' object in descending order (based on confidence).
	 * @param src Integer label of the source Node
	 * @param targ Integer label of the target Node
	 */
	private void sortPathsByConfidence(int src, int targ){
		for(int i=0; i < paths.size(); i++){
			paths.get(i).setConfidence(src, targ);
		} // Set confidence field in each path to avoid re-computation
		Collections.sort(paths);
		return;
	}
	
	/**
	 * Construct the near-optimal DSPG from 'graph', store in 'dspg'.
	 * @param src Integer label of the source Node
	 * @param targ Integer label of the target Node
	 */
	private void findNearOptimalDSPG(int src, int targ){
		dspg = paths.get(0).clone();
		paths.remove(0);
		
		BranchCache bc = new BranchCache();
		Digraph branch = new Digraph();	// Potential addition to DSPG
		boolean endOfPath;				// Whether path is exhausted or not
		Arc nextArc;					// Arc currently being examined
		int position = 0;				// Position in the list of paths
		
		for(int i=0; i < paths.size(); i++){
			position = 0;
			endOfPath = false;
			branch = new Digraph();
			while(!endOfPath){
				nextArc = paths.get(i).arcGet(position);
				position += 1;
				if(position == paths.get(i).arcSize())
					endOfPath = true;
				if(!dspg.hasNode(nextArc.getTarg()))
					branch.addArc(nextArc.clone());
				else if(dspg.hasNode(nextArc.getTarg()) && !branch.isEmpty()){
					branch.addArc(nextArc.clone());
					if(branch.arcSize()<=MAX_CACHE && bc.branchInCache(branch))
						endOfPath = true;
					else if(dspg.sep_subgraph(src, targ, branch)){
						for(int j=0; j < branch.arcSize(); j++)
							dspg.addArc(branch.arcGet(j).clone());
						branch = new Digraph();
					} else
						endOfPath = true;
				} // Check to see if addition will maintain DSPG rules
			} // Consider all branches along a path for inclusion 
		} // Consider all paths in the ordered (by confidence) set
		return;
	}
	
}
