package backend;

import TNA_SL.*;
import interfaces.*;
import java.util.*;

/**
 * The RMDatabase class implements Reputation storage for a network.
 */
public class RMDatabase implements RMDBInterface{
	
	// ************************** PRIVATE FIELDS *****************************
	
	/**
	 * The number of users in this RMDatabase object.
	 */
	private int users;
	
	/**
	 * Storage for all Reputation objects, used vector-wise.
	 */
	private Reputation[][] repStore;
	
	/**
	 * Maps user identifiers (keys?) to indices in this RMDatabase
	 */
	private Map<String, Integer> userLookup;
	
	// *************************** CONSTRUCTORS ******************************
	
	/**
	 * Construct an empty RMDatabase object.
	 */
	public RMDatabase(){
		this.users = 0;
		this.userLookup = new HashMap<String, Integer>();
	}
	
	// ************************ INTERFACED METHODS ***************************
	
	/**
	 * Add information about an interaction between two users.
	 * @param source String ID of user who is submitting the feedback
	 * @param target String ID of user who feedback is being submitted about
	 * @param feedback True to indicate positive feedback; false, otherwise
	 */
	public void addFeedback(String source, String target, boolean feedback){
		int sourceInt = this.getOrAddUser(source);
		int targetInt = this.getOrAddUser(target);
		this.repStore[targetInt][sourceInt].addFeedback(feedback);
	}
	
	/**
	 * Retrieve the number of positive interactions between two users.
	 * @param source String ID of source user of the desired relationship
	 * @param target String ID of target user of the desired relationship
	 * @return Number of positive interactions in [source->target] relationship
	 */
	public int getPos(String source, String target){
		int sourceInt = this.getOrAddUser(source);
		int targetInt = this.getOrAddUser(target);
		return (this.repStore[targetInt][sourceInt].getPos());
	}
	
	/**
	 * Retrieve the number of negative interactions between two users.
	 * @param source String ID of source user of the desired relationship
	 * @param target String ID of target user of the desired relationship
	 * @return Number of negative interactions in [source->target] relationship
	 */
	public int getNeg(String source, String target){
		int sourceInt = this.getOrAddUser(source);
		int targetInt = this.getOrAddUser(target);
		return (this.repStore[targetInt][sourceInt].getNeg());
	}
	
	/**
	 * Retrieve a list of all users contained in this RMDatabase.
	 * @return An array of String objects, each being a unique user ID
	 */
	public String[] getUserList(){
		Set<String> stringSet = this.userLookup.keySet();
		String[] stringArray = new String[stringSet.size()];
		return (stringSet.toArray(stringArray));
	}
	
	// ************************** PUBLIC METHODS *****************************
	
	/**
	 * Export the data in this RMDatabase as a Digraph object.
	 * @return The Digraph representation of this RMDatabase
	 */
	public Digraph exportDigraph(){
		Opinion op;
		Reputation rep;
		Digraph digraph = new Digraph();
		for(int i=0; i < this.users; i++){
			for(int j=0; j < this.users; j++){
				rep = this.repStore[j][i];
				op =  new Opinion(rep.getPos(), rep.getNeg());
				digraph.addArc(new Arc(i, j, op));			
			} // Convert feedbacks to an Opinion object
		} // Opinions then weight the Arcs in a Digraph
		return digraph;
	}
	
	// ************************** PRIVATE METHODS ****************************
	
	/**
	 * Determine the index of a particular user, if it DNE, create it.
	 * @param userID String identifier of the user index we are determining
	 * @return The index of 'userID' in this RMDatabase object
	 */
	private int getOrAddUser(String userID){
		if(this.userLookup.containsKey(userID))
			return this.userLookup.get(userID);
		else
			return (this.addUser(userID));
	}
	
	/**
	 * Increase the capacity of this RMDatabase object by one user.
	 * @param userID Unique string id (key?) of the user being added 
	 * @return The index of the new user that was created
	 */
	private int addUser(String userID){
		this.users++;
		int lastIndex = (this.users - 1);
		Reputation[][] newStore = new Reputation[this.users][this.users];
		for(int i=0; i < this.users; i++){
			for(int j=0; j < this.users; j++){
				if(i == lastIndex || j == lastIndex)
					newStore[i][j] = new Reputation();
				else
					newStore[i][j] = this.repStore[i][j];
			} // Reputations involving the new user get a new Reputation
		} // Just copy all Reputations not involving the new user
		this.repStore = newStore;
		this.userLookup.put(userID, lastIndex);
		return lastIndex;
	}
	
}
