package backend;

import java.util.*;

/**
 * The Reputation class defines the relationship between two users, storing
 * directed information about past interaction from one users perspective.
 */
public class Reputation{
	
	// ************************** PRIVATE FIELDS *****************************
	
	/**
	 * Stores the number of positive prior interactions.
	 */
	private int posFeedbacks;
	
	/**
	 * Stores the number of negative prior interactions.
	 */
	private int negFeedbacks;
	
	/**
	 * Determines whether a feedback history should be maintained.
	 */
	private boolean historyEnabled; 
	
	/**
	 * A feedback history, whose use is determined by boolean 'historyEnabled'.
	 */
	private List<Boolean> feedbackHistory;
	
	// *************************** CONSTRUCTORS ******************************
	
	/**
	 * Construct a default, empty Reputation object with history enabled. 
	 */
	public Reputation(){
		this.posFeedbacks = 0;
		this.negFeedbacks = 0;
		this.historyEnabled = true;
		this.feedbackHistory = new ArrayList<Boolean>();
	}
	
	/**
	 * Construct a Reputation object, with history preference as an argument.
	 * @param enableHistory Whether or not f-back history should be maintained.
	 */
	public Reputation(boolean enableHistory){
		this.posFeedbacks = 0;
		this.negFeedbacks = 0;
		this.historyEnabled = enableHistory;
		if(this.historyEnabled)
			this.feedbackHistory = new ArrayList<Boolean>();
	}
	
	// ************************** PUBLIC METHODS *****************************
	
	/**
	 * Add information about an interaction to this Reputation object .
	 * @param feedback True if the interaction was positive, false otherwise
	 */
	public void addFeedback(boolean feedback){
		if(feedback)
			this.posFeedbacks++;
		else
			this.negFeedbacks++;
		if(this.historyEnabled)
			this.feedbackHistory.add(Boolean.valueOf(feedback));
		return;
	}
	
	/**
	 * Calculate the 'feedback integer' for this Reputation object.
	 * @return The integer (pos-neg) if it is greater than 0; 0 otherwise.
	 */
	public int calculateFeedbackInt(){
		int fback_int = (this.getPos() - this.getNeg());
		if(fback_int < 0)
			fback_int = 0;
		return fback_int;	
	}
	
	/**
	 * Return the # of positive interactions characterizing this relationship.
	 * @return the # of positive interactions characterizing this relationship
	 */
	public int getPos(){
		return this.posFeedbacks;
	}
	
	/**
	 * Return the # of negative interactions characterizing this relationship.
	 * @return the # of negative interactions characterizing this relationship
	 */
	public int getNeg(){
		return this.negFeedbacks;
	}

}
