
package quantify;

import interfaces.*;

/**
 * The Quantifier class allows us to combine three types of data; 1) A trust
 * dependency graph from the KeyNote interpreter. 2) Reputation data. and 
 * 3) A Reputation management algorithm. The output is the trust value one
 * should associate with the delegation/credential. 
 */
public class Quantifier{
	
	// *************************** PUBLIC FIELDS *****************************
	
	/**
	 * The Strategy enumeration lists varying evaluation methods.
	 * For example, say the delegation is (a->b->c).
	 * 'multiply_hops': Will yield [trust(a->b)*trust(b->c)]
	 * 'single_hop': Will yield [trust(a->c)]
	 * 'minimum': Will yield [minimum(trust(a->b), trust(b->c))]
	 */
	public enum Strategy{multiply_hops, single_hop, minimum};
	
	// ************************** PRIVATE FIELDS *****************************
	
	/**
	 * The TDPath object detailing delegation in a credential.
	 */
	private TDPath tdp;
	
	// *************************** CONSTRUCTORS ******************************
	
	/**
	 * Construct a Quantifier object by providing the TDPath to be evaluated.
	 * @param trustPath TDPath object for evaluation (describes delegation)
	 */
	public Quantifier(TDPath trustPath){
		this.tdp = trustPath;
	}
	
	// ************************** PUBLIC METHODS *****************************
	
	/**
	 * Combine reputation data and a reputation algorithm according to a
	 * provided 'strategy' to produce a trust value for 'this' TDP.
	 * @param rDB Reputation store containing relevant feedback data
	 * @param tm Instance of TM-algorithm to be used in quantification
	 * @param strat Quantification strategy. See enumeration for descriptions.
	 * @return Trust value on [0..1], or NaN if it can't be computed
	 */
	public double quantify(RMDBInterface rDB, TMInterface tm, Strategy strat){
		double trustValue;
		if(strat == Strategy.single_hop)
			trustValue = quantifySingle(rDB, tm);
		else if(strat == Strategy.multiply_hops)
			trustValue = quantifyMultiply(rDB, tm);
		else // Strategy.minimum
			trustValue = quantifyMinimum(rDB, tm);
		return trustValue;
	}
	
	// ************************** PRIVATE METHODS ****************************
	
	/**
	 * Use the 'single_hop' quantification method to measure trust. 
	 * @param rDB Reputation store containing relevant feedback data
	 * @param tm Instance of TM-algorithm to be used in quantification
	 * @return Trust value on [0..1], or NaN if it can't be computed
	 */
	private double quantifySingle(RMDBInterface rDB, TMInterface tm){
		String source = tdp.getStart();
		String target = tdp.getEndNode();
		tm.importRMData(rDB);
		return (tm.computeTrust(source, target));
	}
	
	/**
	 * Use the 'multiply_hops' quantification method to measure trust.
	 * @param rDB Reputation store containing relevant feedback data
	 * @param tm Instance of TM-algorithm to be used in quantification
	 * @return Trust value on [0..1], or NaN if it can't be computed
	 */
	private double quantifyMultiply(RMDBInterface rDB, TMInterface tm){
		
			// Prepare the TM algorithm, investigate first hop
		tm.importRMData(rDB);
		String source = tdp.getStart();
		String target = tdp.getTarget(source);
		
			// If an arc error exists, then invalid delegation; return error 
		if(tm.computeTrust(source, target) == Double.NaN)
			return (Double.NaN);
		
			// Multiply trust along all hops until end-point reached
		double trustValue = 1.0, tempTrust;
		while(target != null){
			tempTrust = tm.computeTrust(source, target);
			if(tempTrust == Double.NaN)
				return (Double.NaN);
			trustValue *= tempTrust;
			source = target;
			target = tdp.getTarget(source);
		} // Follow path until the end 		
		return (trustValue);
	}
	
	/**
	 * Use the 'minimum' quantification method to measure trust.
	 * @param rDB Reputation store containing relevant feedback data
	 * @param tm Instance of TM-algorithm to be used in quantification
	 * @return Trust value on [0..1], or NaN if it can't be computed
	 */
	private double quantifyMinimum(RMDBInterface rDB, TMInterface tm){
		
			// Prepare the TM algorithm, investigate first hop
		tm.importRMData(rDB);
		String source = tdp.getStart();
		String target = tdp.getTarget(source);
		
			// If an arc error exists, then invalid delegation; return error 
		if(tm.computeTrust(source, target) == Double.NaN)
			return (Double.NaN);
		
			// Compute minimum of all hops until end-point reached
		double minimum = Double.MAX_VALUE, tempTrust;
		while(target != null){
			tempTrust = tm.computeTrust(source, target);
			if(tempTrust == Double.NaN)
				return (Double.NaN);
			minimum = Math.min(minimum, tempTrust);
			source = target;
			target = tdp.getTarget(source);
		} // Follow path to its conclusion		
		return (minimum);
	}
	
}
