
package quantify;

import java.util.*;

/**
 * The TDPath class implements a trust-dependency-path (TDP), a subset of what
 * may eventually be a trust-dependency graph (TDG). 
 */
public class TDPath{
	
	// ************************** PRIVATE FIELDS *****************************
	
	/**
	 * String identifier of the root node in this TDPath.
	 */
	private String startNode;
	
	/**
	 * Mapping of source nodes to target nodes in this TDPath.
	 */
	private Map<String, String> arcs;

	// *************************** CONSTRUCTORS ******************************
	
	/**
	 * Construct a TDPath object by providing the root identifier.
	 * @param start Unique string ID of the root node for this TDPath
	 */
	public TDPath(String start){
		this.startNode = start;
		this.arcs = new HashMap<String, String>();
	}
	
	// ************************** PUBLIC METHODS *****************************
	
	/**
	 * Add an arc to this TDPath
	 * @param source Unique string ID of the source node for this arc
	 * @param target Unique string ID of the target node for this arc
	 */
	public void addArc(String source, String target){
		this.arcs.put(source, target);
	}
	
	/**
	 * Give a source node, retrieve which target node that source points too.
	 * @param source Unique string ID of a source node
	 * @return String identifier of the 'target'; null if one DNE
	 */
	public String getTarget(String source){
		return (this.arcs.get(source));
	}
	
	/**
	 * Return the 'startNode' field of this TDPath object.
	 * @return the 'startNode' field of this TDPath object.
	 */
	public String getStart(){
		return (this.startNode);
	}
	
	/**
	 * Find the end (terminal) node of this TDPath.
	 * @return Last node reachable from 'startNode' via path traversal
	 */
	public String getEndNode(){
		String source = this.startNode;
		String target = this.getTarget(source);
		while(target != null){
			source = target;
			target = this.getTarget(source);
		} // Follow path until the end is reached
		return source;
	}

}
